
/*!
 * Overrides Jade to give it i18n capabilities.
 */

var jade = module.exports = require('jade')
  , crypto = require('crypto')
  , nodes = jade.nodes
  , runtime = jade.runtime
  , jadeLexer = jade.Lexer
  , jadeParser = jade.Parser
  , jadeCompiler = jade.Compiler
  , EventEmitter = require('events').EventEmitter

/**
 * i18n API.
 */

var i18n = jade.i18n = new EventEmitter;
i18n.hash = hash;
i18n.trimPhrase = trimPhrase;
i18n.phrases = {};

/**
 * Version.
 */

jade.i18n.version = '0.4.4';

/**
 * Overrides.
 */

jade.Compiler = Compiler;
jade.Parser = Parser;

/**
 * Parse the given `str` of jade and return a function body.
 
 * @param {String} str
 * @param {Object} options
 * @return {String}
 * @api private
 */

function parse(str, options){
  try {
    // Parse
    var parser = new Parser(str, options.filename, options);

    // Compile
    var compiler = new (options.compiler || Compiler)(parser.parse(), options)
      , js = compiler.compile();

    // Debug compiler
    if (options.debug) {
      console.error('\nCompiled Function:\n\n\033[90m%s\033[0m', js.replace(/^/gm, '  '));
    }

    return ''
      + 'var buf = [];\n'
      + (options.self
        ? 'var self = locals || {};\n' + js
        : 'with (locals || {}) {\n' + js + '\n}\n')
      + 'return buf.join("");';
  } catch (err) {
    parser = parser.context();
    runtime.rethrow(err, parser.filename, parser.lexer.lineno);
  }
}

/**
 * Override jade#compile so that we can override the Lexer, Parser and add the
 * `_` local.
 *
 * @api public
 */

jade.compile = function(str, options){
  var options = options || {}
    , client = options.client
    , filename = options.filename
      ? JSON.stringify(options.filename)
      : 'undefined'
    , fn;

  if (options.compileDebug !== false) {
    fn = [
        'var __ = [{ lineno: 1, filename: ' + filename + ' }];'
      , 'try {'
      , parse(String(str), options || {})
      , '} catch (err) {'
      , '  rethrow(err, __[0].filename, __[0].lineno);'
      , '}'
    ].join('\n');
  } else {
    fn = parse(String(str), options || {});
  }

  if (client) {
    fn = 'var attrs = jade.attrs, escape = jade.escape, rethrow = jade.rethrow;\n' + fn;
  }

  fn = new Function('locals, attrs, escape, rethrow, _', fn);

  if (client) return fn;

  return function(locals){
    _.language = options.language;
    _.context = { name: '', default: true };
    return fn(locals, runtime.attrs, runtime.escape, runtime.rethrow, _);
  };
};

/**
 * Lexer constructor.
 *
 * @api public
 */

function Lexer (str, options) {
  jadeLexer.call(this, str, options);
  this.lastTok = null;
}

/**
 * Inherits from Lexer.
 */

Lexer.prototype.__proto__ = jadeLexer.prototype;

/**
 * Translatable token detection (+)
 *
 * @api private
 */

Lexer.prototype.translatable = function () {
  if ('tag' == this.lastTok
    || 'id' == this.lastTok
    || 'class' == this.lastTok
    || 'attrs' == this.lastTok) {
    return this.scan(/^\+/, 'translatable');
  }
};

/**
 * Override attrs to transform `i` into `i18n`
 *
 * @api private
 */

Lexer.prototype.attrs = function () {
  var tok = jadeLexer.prototype.attrs.call(this);
  if (tok) {
    if (tok.attrs.i) {
      tok.attrs.i18n = tok.attrs.i;
      delete tok.attrs.i;
    }

    return tok;
  }
};

/**
 * Keep track of the last token to disambiguate.
 *
 * @api private
 */

Lexer.prototype.tok = function (type, val) {
  this.lastTok = type;
  return jadeLexer.prototype.tok.call(this, type, val);
};

/**
 * Extend `next` to add i18n tokens probing.
 */

Lexer.prototype.next = function () {
  return this.translatable() || jadeLexer.prototype.next.call(this);
};

/**
 * Parser constructor.
 *
 * @api public
 */

function Parser (str, filename, options) {
  jadeParser.call(this, str, filename, options);
  this.lexer = new Lexer(str, options);
}

/**
 * Inhertis from Parser.
 */

Parser.prototype.__proto__ = jadeParser.prototype;

/**
 * Parse tag. Identical to the original, except we look for a translatable
 * token in the token loop.
 *
 * @api private
 */

var textOnly = jadeParser.textOnly;

Parser.prototype.parseTag = function () {
  // ast-filter look-ahead
  var i = 2;
  if ('attrs' == this.lookahead(i).type) ++i;
  if (':' == this.lookahead(i).type) {
    if ('indent' == this.lookahead(++i).type) {
      return this.parseASTFilter();
    }
  }

  var name = this.advance().val
    , tag = new nodes.Tag(name)
    , dot;

  tag.line = this.line();

  // (attrs | class | id)*
  out:
    while (true) {
      switch (this.peek().type) {
        // add empty i18n atttribute
        case 'translatable':
          var tok = this.advance();
          tag.setAttribute('i18n', '""');
          continue;
        case 'id':
        case 'class':
          var tok = this.advance();
          tag.setAttribute(tok.type, "'" + tok.val + "'");
          continue;
        case 'attrs':
          var obj = this.advance().attrs
            , names = Object.keys(obj);
          for (var i = 0, len = names.length; i < len; ++i) {
            var name = names[i]
              , val = obj[name];
            tag.setAttribute(name, val);
          }
          continue;
        default:
          break out;
      }
    }

  // check immediate '.'
  if ('.' == this.peek().val) {
    dot = tag.textOnly = true;
    this.advance();
  }

  // (text | code | ':')?
  switch (this.peek().type) {
    case 'text':
      tag.text = this.parseText();
      break;
    case 'code':
      tag.code = this.parseCode();
      break;
    case ':':
      this.advance();
      tag.block = new nodes.Block;
      tag.block.push(this.parseTag());
      break;
  }

  // newline*
  while ('newline' == this.peek().type) this.advance();

  tag.textOnly = tag.textOnly || ~textOnly.indexOf(tag.name);

  // script special-case
  if ('script' == tag.name) {
    var type = tag.getAttribute('type');
    if (!dot && type && 'text/javascript' != type.replace(/^['"]|['"]$/g, '')) {
      tag.textOnly = false;
    }
  }

  // block?
  if ('indent' == this.peek().type) {
    if (tag.textOnly) {
      this.lexer.pipeless = true;
      tag.block = this.parseTextBlock();
      this.lexer.pipeless = false;
    } else {
      var block = this.block();
      if (tag.block) {
        for (var i = 0, len = block.nodes.length; i < len; ++i) {
          tag.block.push(block.nodes[i]);
        }
      } else {
        tag.block = block;
      }
    }
  }

  return tag;
};

/**
 * Override parseExpr for `+` when used with class/id
 *
 * @api private
 */

Parser.prototype.parseExpr = function () {
  if ('translatable' == this.peek().type &&
    'tag' == this.lexer.deferredTokens[0].type) {
    this.lexer.defer(this.advance());
  }

  return jadeParser.prototype.parseExpr.call(this);
};

/**
 * Overrid parseInclude to reference our parser.
 */

Parser.prototype.parseInclude = function(){
  var path = require('path')
    , fs = require('fs')
    , dirname = path.dirname
    , basename = path.basename
    , join = path.join;

  var path = name = this.expect('include').val.trim()
    , dir = dirname(this.filename);

  if (!this.filename)
    throw new Error('the "filename" option is required to use includes');

  // non-jade
  if (~basename(path).indexOf('.')) {
    var path = join(dir, path)
      , str = fs.readFileSync(path, 'utf8');
    return new nodes.Literal(str);
  }

  var path = join(dir, path + '.jade')
    , str = fs.readFileSync(path, 'utf8')
   , parser = new Parser(str, path, this.options);

  this.context(parser);
  var ast = parser.parse();
  this.context();
  ast.filename = path;

  if ('indent' == this.peek().type) {
    ast.lastBlock().push(this.block());
  }

  return ast;
};

/**
 * Compiler constructor
 *
 * @api public
 */

function Compiler (node, options) {
  jadeCompiler.call(this, node, options);
  this.i18nContext = { name: '', default: true };
}

/**
 * Inherits from Compiler.
 */

Compiler.prototype.__proto__ = jadeCompiler.prototype;

/**
 * Performs replacement of text for i18n.
 *
 * @param {Node} node
 * @param {i18n} context, optional
 * @api private
 */

Compiler.prototype.visitNode = function (node) {
  var parent = jadeCompiler.prototype.visitNode;

  if ('i18n' == node.name) {
    if (node.block.nodes.length) {
      throw new Error('Tag "i18n" cannot contain nested elements');
    } else if ('' == node.text.nodes[0].trim()) {
      throw new Error('Tag "i18n" must contain text');
    }

    var context = { name: node.text.nodes[0].trim() };

    node.attrs.forEach(function (attr) {
      context[attr.name] = 'string' == typeof attr.val 
        ? attr.val.replace(/^['"]|['"]$/g, '')
        : attr.val
    });

    // check if its the first defined context
    var firstContext = true == this.i18nContext.default;

    if (node.parent) {
      node.parent.i18nContext = context;
    } else {
      this.i18nContext = context;
    }

    // only set _.context for the main file context
    if (!firstContext) return;

    // replace node with an inline representation so that it can be accessed by `_`
    node = new nodes.Code('_.context = ' + JSON.stringify(context), false, false);
  } else if (node.getAttribute && undefined !== node.getAttribute('i18n')) {
    // get phrase data
    var text = node.textOnly ? node.block.nodes.join('') : node.text.nodes[0]
      , desc = node.attrs.filter(function (attr) {
          return 'i18n' == attr.name;
        })[0] || null

    if (desc) {
      desc = desc.val.replace(/^['"]|['"]$/g, '');
    }

    var context = this.getI18nContext(node)
      , h = i18n.hash(text, desc, context)

    // emit `translatable phrase` event
    jade.i18n.emit('translatable phrase', {
        hash: h
      , phrase: trimPhrase(text)
      , description: desc
      , context: context
    });

    if (this.options.language) {
      if (node.block.nodes.length && !node.textOnly) {
        throw new Error("Translatable items can't have children");
      }

      // look up translation

      var lang = this.options.language
        , langTranslations = i18n.phrases[lang] || {}
        , translation = langTranslations[h]

      if (translation) {
        if (node.textOnly) {
          translation = translation
            .replace(/\r/g, '')
            .replace(/\n/g, '\r\n\r')
            .split('\r')
            .map(function (v) {
              return v.replace(/\n/g, '\\n');
            });

          // clean up translators whitespace
          translation[translation.length - 1] = translation[translation.length - 1]
            .replace(/ *$/g, '');

          // apply consistent whitespace
          if (text.length > 1 && ' ' == text[text.length - 1]) {
            translation[translation.length - 1] += ' ';
          }

          node.block.nodes = translation;
        } else {
          // trim translator's whitespace
          translation = translation.trim();

          // apply consistent whitespace
          if (text.length > 1 && ' ' == text[text.length - 1]) {
            translation += ' ';
          }

          node.text.nodes[0] = translation;
        }
      } else {
        jade.i18n.emit('missing translation', {
            lang: lang
          , hash: h
          , phrase: trimPhrase(text)
          , description: desc
          , context: context
        });
      }
    }

    // remove i18n attr
    node.removeAttribute('i18n');
  } else if (node.block && node.block.nodes && node.block.nodes.length) {
    for (var i = 0, l = node.block.nodes.length; i < l; i++) {
      node.block.nodes[i].parent = node;
    }
  }

  return parent.call(this, node);
};

/**
 * Retrieves a context for a certain node.
 *
 * @param {Node} node to get context for
 * @api private
 */

Compiler.prototype.getI18nContext = function (node) {
  var parent = node.parent;

  while (parent) {
    if (parent.i18nContext) {
      return parent.i18nContext;
    }
    parent = parent.parent;
  }

  return this.i18nContext;
};

/**
 * Helper for runtime jade translations (_).
 *
 * @param {String} phrase text
 * @param {String} optional, description
 * @param {Object} optional, substitutions
 * @api private
 */

function _(text, desc, substitutions) {
  if ('object' == typeof desc) {
    substitutions = desc;
    desc = null;
  }

  var context = _.context || { name: '', default: true }
    , h = i18n.hash(text, desc, context)
    , lookup

  jade.i18n.emit('translatable phrase', {
      hash: h
    , phrase: trimPhrase(text)
    , description: desc || ''
    , context: context
  });

  if (_.language) {
    if (i18n.phrases[_.language]) {
      lookup = i18n.phrases[_.language][h];
    }

    if (undefined == lookup) {
      jade.i18n.emit('missing translation', {
          lang: _.language
        , hash: h
        , phrase: trimPhrase(text)
        , description: desc
        , context: context
      });
    }
  }

  if (substitutions) {
    return substitute(undefined != lookup ? lookup : text, substitutions);
  } else {
    return lookup || text;
  }
}

/**
 * Performs substitutions in a string.
 * Example:
 *    substitute('Hello {name}', {name: "woot"})
 *
 * @param {String} str
 * @param {Object} hash of replacements.
 * @api private
 */

function substitute (str, object) {
  return str.replace(/\\?\{([^{}]+)\}/g, function (match, name) {
    if (match.charAt(0) == '\\') return match.slice(1);
    return (object[name] != undefined) ? object[name] : '';
  });
}

/**
 * Hashes a phrase.
 *
 * @param {String} phrase
 * @param {String} description
 * @param {Object} context
 * @api private
 */

function hash (phrase, description, context) {

  var h = trimPhrase(phrase)
    + '|☃|' + (description || '').replace(/ +/g, ' ').trim()
    + '|☃|' + (context.name || '').replace(/ +/g, ' ').trim();

  return h.toLowerCase();
}

/**
 * Trims a phrase.
 *
 * @param {String} phrase
 * @api private
 */

function trimPhrase (phrase) {
  // convert \\n to \n (textonly)
  phrase = phrase.replace(/\\n/g, '\n');

  // remove double spaces and unneeded characters
  phrase = phrase.replace(/\s+/g, ' ');

  // trim
  phrase = phrase.trim();

  return phrase;
};
